;---------------------------------------------------------------------
;  	A Certified implementation of the                          
;	 	                                  
;	I M P E R A T I V E  R H O  C A L C U L U S                  
;                                   
;       Powered by Bigloo-Scheme                      
;                                   
;       Copyright (c) INRIA 2004-2005 
;       Luigi Liquori and Bernard Paul Serpette 
;                                   
;  	Bug descriptions, use reports,                         
;	comments or suggestions are                            
;  	welcome. Send them to                            
;	                                  
;	Luigi.Liquori@inria.fr                             
;       Bernard.Serpette@inria.fr                            
;                  
;Copyright (C) 2004-2005 Luigi Liquori, Bernard Paul Serpette, INRIA.
;
;This library is free software; you can redistribute it and/or
;modify it under the terms of the GNU Lesser General Public
;License as published by the Free Software Foundation; either
;version 2.1 of the License, or (at your option) any later version.
;
;This library is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;Lesser General Public License for more details.
;
;You should have received a copy of the GNU Lesser General Public
;License along with this library; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;-----------------------------------------
;;
;;
;;
;;
;;
;; MODULE DECLARATIONS AND DATATYPES
;;
(module imprho
	(main interpreter)
	(static
	 (class Expr)
	 (class Alg::Expr      value::symbol)
	 (class Var::Expr       name::symbol)
	 (class Abs::Expr       patt::Expr body::Expr)
	 (class App::Expr        fun::Expr arg::Expr)
	 (class StructE::Expr   left::Expr right::Expr)
	 (class Ref::Expr       expr::Expr)
	 (class Deref::Expr     expr::Expr)
	 (class Assign::Expr    left::Expr right::Expr)
	 (class Msg::Expr        msg::string)
	 (class Exit::Expr       msg::string)
;;
;;
	 (class Value)
	 (class Lit::Value     value::symbol)
	 (class Call::Value      fun::Value arg::Value)
	 (class Clos::Value    funct::Abs env)
	 (class StructV::Value  left::Value right::Value)
	 (class Loc::Value      addr::int)
	 (class Wrong::Value     msg::string)
	 (class MsgV::Value      msg::string)
	 (class ExitV::Value     msg::string)))
;;
;; LOADING NOSTUCK THEORY
;;
(define *cleaning_wrong*  #t)
;;                                   
;; READ EVAL PRINT LOOP                           
;;                              
(define
   (interpreter argv)   
   (begin
    (print " ----------------------------------------------")
    (print "|               ----------\\                    |")
    (print "|               | i R h o  -                   |")
    (print "|               ----------/                    |")
    (print "| An Imperative Rewriting Calculus Interpreter |")
    (print "|     Certified by the Proof Assistant Coq     |")
    (print "|           Powered by Bigloo Scheme           |")
    (print "|             Copyright Inria 2005             |")
    (print "|             Nostuck Theory Loaded            |")
    (print "|             $ = Switch Theory                |")
    (print "|             # = Clean  Namespace             |")
    (print "|             @ = Exit   iRho                  |")
    (print " ---------------------------------------------- ")
;
;
    (let loop ()
       (bind-exit (return)
	  (with-exception-handler
	   (lambda (exc) 
	      (if (&error? exc)
		  (print "(ERROR: Bad Parsing on " (&error-obj exc) ")")
		  (print "(ERROR: " exc ")"))
	      (return #t))
	   (lambda ()
	      (init_store)
	      (display* "iRho  IN > ")
	      (let ((val (Eval (parse (current-input-port)) *namespace*)))
		 (print "iRho OUT > " (extern val))))))
       (loop))))
;;
;;
;; DEFINE WRONG BY SEGMENTATION VIOLATION
;;
(define *segment_fault*
  (lambda (loc)
     (instantiate::Wrong
	(msg (string-append "Segmentation Violation on Location "
			    (integer->string (Loc-addr loc)))))))
;;
;;
;; DEFINE WRONG BY UNBOUND VARIABLE
;;
(define *unbound_var*
   (lambda (var)
      (instantiate::Wrong
	 (msg (string-append "Unbound Variable "
			     (symbol->string (Var-name var)))))))
;;
;;
;; DEFINE A STORE AN ADDRESS_STORE_COUNTER AND AN INIT_STORE
;;
(define *store* *segment_fault*)
;;
(define *address_store_counter* 0)
;;
(define (init_store)
  (set! *address_store_counter* 0)
  (set! *store* *segment_fault*))
;;
;;
;;
;; DEFINE A NAMESPACE, AN INIT_NAMESPACE 
;;
;;
(define *namespace* *unbound_var*)
;;
(define (init_namespace)
   (set! *namespace* *unbound_var*))
;;
;;
;; ADD AN IDENTIFIER IN A GLOBAL NAMESPACE
;;
(define (add_namespace ident expression)
   (let ((val     (Eval expression *namespace*))
	 (namesp  *namespace*))
      (set! *namespace* (lambda (v) (if (eq? (Var-name v) ident) val (namesp v))))))
;;
;;
;;
;; CREATE NEW LOCATIONS IN A STORE 
;;
(define (new_location val::Value)
   (let ((newloc (instantiate::Loc (addr *address_store_counter*))))
     (set! *address_store_counter* (+ *address_store_counter* 1))
     (let ((old_store *store*))
       (set! *store* (lambda (loc) (if (eq? loc newloc) val (old_store loc)))))
     newloc))
;;
;;
;;
;;                                  
;; INPUT                              
;;                                  
(define-generic (code expr::Expr))
;;                          
(define-method (code expr::Alg) 
  (Alg-value expr))
;;
(define-method (code expr::Var)                         
  (Var-name expr))
;;
(define-method (code expr::Abs)
   (list 'CLOSENV '< (code (Abs-patt expr)) '-> (code (Abs-body expr)) '>))
;;
(define-method (code expr::App)                         
  (list (code (App-fun expr)) (code (App-arg expr))))
;;
(define-method (code expr::StructE)                         
  (list (code (StructE-left expr))
	","
	(code (StructE-right expr))))
;;                      
(define-method (code expr::Ref)                         
  (list 'REF (code (Ref-expr expr))))
;;
(define-method (code expr::Deref)                         
  (list 'BANG (code (Deref-expr expr))))
;;
(define-method (code expr::Assign)                         
  (list 'ASS
	(code (Assign-left expr))
	(code (Assign-right expr))))
;;
(define-method (code expr::Msg)
  (Msg-msg  expr))
;;
(define-method (code expr::Exit)
  (Exit-msg  expr))
;;

;;
;;
;; OUTPUT                             
;;                                  
(define-generic (extern value::Value))                        
;;                                  
(define-method (extern value::Lit)                         
  (Lit-value value))
;;                               
(define-method (extern value::Call)                          
  (let ((f (extern (Call-fun value)))
	(a (extern (Call-arg value))))
    (if (pair? f) (append f (cons a '())) (list f a))))
;;                                 
(define-method (extern value::Clos)                                  
  (code (Clos-funct value)))
;;                                  
(define-method (extern value::StructV)                         
  (list (extern (StructV-left value)) "," (extern (StructV-right value))))
;;                                  
(define-method (extern value::Loc)                         
 (list (string-append
	"LOC "	
	(integer->string (Loc-addr value))
	" =")
       (extern (*store* value))))
;;                                  
(define-method (extern value::Wrong)                         
  (string-append "(WRONG: " (Wrong-msg value) ")"))
;;
(define-method (extern value::MsgV)                         
  (MsgV-msg value))
;;
(define-method (extern value::ExitV)                         
  (print (ExitV-msg value))
  (exit 0))



;;                                
;;
;;
;;
;;
;;
;; 
;; S T A R T  O F  T H E  C E R T I F I E D  C O D E
;;
;;
;;
;;
;;
;; EVAL FUNCTION                               
;;                                  
(define-generic (Eval::Value expr::Expr env))                              
;;                                                         
(define-method (Eval::Value expr::Alg env)                         
  (instantiate::Lit (value (Alg-value expr))))
;;                                  
(define-method (Eval::Value expr::Var env)                       
  (let ((inenv (env expr)))
     (if (Wrong? inenv)
	 (*namespace* expr)
	 inenv)))
      
     
;;                                  
(define-method (Eval::Value expr::Abs env)                         
  (instantiate::Clos (funct expr) (env env)))
;;                                  
(define-method (Eval::Value expr::App env)                         
  (let   ((fun (Eval (App-fun expr) env)))
    (let ((arg (Eval (App-arg expr) env)))
      (call fun arg))))
;;                                  
(define-method (Eval::Value expr::StructE env)                         
  (let   ((left  (Eval (StructE-left  expr) env)))
    (let ((right (Eval (StructE-right expr) env)))
      (instantiate::StructV (left  left)
			    (right right)))))
;;                                  
(define-method (Eval::Value expr::Ref env)     
  (new_location (Eval (Ref-expr expr) env)))
;;                                  
(define-method (Eval::Value expr::Deref env)                         
  (let ((loc (Eval (Deref-expr expr) env)))
    (if (Loc? loc)
	(*store* loc)
      (instantiate::Wrong (msg "Deref is Not a Location")))))
;;                                  
(define-method (Eval::Value expr::Assign env)                         
  (let ((val (Eval (Assign-right expr) env)))
    (let ((loc (Eval (Assign-left expr) env)))
      (if (Loc? loc)
	  (let ((old *store*))
	    (set! *store* (lambda (x) (if (eq? x loc) val (old x))))
	    val)
	(instantiate::Wrong (msg "Assign is not a location"))))))
;;    
(define-method (Eval::Value expr::Msg env)                         
  (instantiate::MsgV (msg (Msg-msg expr))))
;;
(define-method (Eval::Value expr::Exit env)                         
   (instantiate::ExitV (msg (Exit-msg expr))))
;;                              
;;
;;
;;                   
;; CALL FUNCTION                                 
;;                                  
(define-generic (call::Value fun::Value arg::Value))                     
;;
(define-method (call::Value fun::Value arg::Value)  
  (instantiate::Call (fun fun) (arg arg)))
;;
(define-method (call::Value fun::Clos arg::Value)                         
  (let ((newenv (match (Abs-patt (Clos-funct fun)) arg (Clos-env fun))))
    (if newenv 
 	(Eval (Abs-body (Clos-funct fun))  newenv)
      (instantiate::Wrong (msg "Pattern Mismatch")))))
;;                          
(define-method (call::Value fun::StructV arg::Value)                         
  (if *cleaning_wrong*
      (let ((val1 (call (StructV-left  fun) arg)) 
	    (val2 (call (StructV-right fun) arg)))
	(cond ((Wrong? val1) val2)
	      ((Wrong? val2) val1)
	      (else (instantiate::StructV (left val1) (right val2)))))
    (instantiate::StructV (left  (call (StructV-left  fun) arg)) 
			  (right (call (StructV-right fun) arg)))))
;;                                  
(define-method (call::Value fun::Loc arg::Value)
  (instantiate::Wrong (msg "Calling a Location")))
;;                                  
(define-method (call::Value fun::Wrong arg::Value)                         
  fun)
;;    
;;
;;                              
;;
;;
;;
;;
;; MATCH FUNCTION
;;
(define-generic (match patt::Expr val::Value env))
;;                                  
(define-method (match patt::Alg val::Value env)                         
  (if (and (Lit? val)
	   (equal? (Lit-value val) (Alg-value patt)))
      env
    #f))
;;
(define-method (match patt::Var val::Value env)
   (let ((oldval (env patt))) 
      (cond
	 ((Wrong? oldval)     (lambda (x) (if (eq? (Var-name x) (Var-name patt)) val (env x))))
	 ((equal? oldval val) env)
	 (else                #f))))
;;                                  
(define-method (match patt::App val::Value env)
  (if (Call? val)
       (let ((newenv (match (App-fun patt) (Call-fun val) env)))
	 (if  newenv (match (App-arg patt) (Call-arg val) newenv) #f))
    #f))
;;              
(define-method (match patt::StructE val::Value env)   
   (if (StructV? val)
       (let ((newenv (match (StructE-left  patt) (StructV-left  val) env)))
          (if newenv (match (StructE-right patt) (StructV-right val) newenv) #f))
       #f))
;;           
(define-method (match patt::Ref val::Value env)                         
  (if (Loc? val)
      (match (Ref-expr patt) (*store* val) env)
    #f))
;;                                  
(define-method (match patt::Deref val::Value env)                         
  #f)
;;                                  
(define-method (match patt::Assign val::Value env)                         
  #f)
;;        
(define-method (match patt::Abs val::Value env) 
  #f)           
;;        
;;
;;
;;
;;
;;
;; 
;; E N D  O F  T H E  C E R T I F I E D  C O D E
;;
;;
;;
;;
;;
;; THE PARSER                               
;;             
;;
;; CURRIFICATION FUNCTIONS
;;
(define (curry list)
  (if (null? list) (instantiate::Alg (value list))
    (curry2 (car list) (cdr list))))
;;  
(define (curry2 fun list)
  (if (null? list)
      fun
    (curry2
     (instantiate::App (fun fun) (arg (car list)))
     (cdr list))))
;;
;;
;;
;;
;; SET!FIX
;;
(define (set!fix ident expression)
  (let ((ident_ref   (symbol-append ident 'REF)))
    (begin 
     (add_namespace ident        (compilefix ident ident_ref expression))
     (add_namespace ident_ref    (instantiate::Var (name ident))))))
;;
;;
;;
;;
;; COMPILEFIX
;;
(define (compilefix ident ident_ref expression)
  (instantiate::App 
   (fun (instantiate::Abs
	 (patt (instantiate::Var (name ident_ref)))
	 (body (instantiate::App 
		(fun (instantiate::Abs  (patt (instantiate::Var (name ident)))
					(body (instantiate::Assign 
					       (left  (instantiate::Var (name ident_ref)))
					       (right (instantiate::Var (name ident)))))))
		(arg (replace 
		      ident 
		      (instantiate::Deref   (expr  (instantiate::Var (name ident_ref))))
		      expression))))))
   (arg (instantiate::Ref (expr (instantiate::Alg (value 'dummy)))))))
;;
;;
;;
;;
;; REPLACE FUNCTION
;;
(define (replace id id_ref ex::Expr)
  (cond 
   ((Alg?     ex)   ex)
   ((Var?     ex)   (if (eq? id (Var-name ex)) id_ref ex))
   ((App?     ex)   (instantiate::App     (fun   (replace id id_ref (App-fun       ex)))
					  (arg   (replace id id_ref (App-arg       ex)))))
   ((Assign?  ex)   (instantiate::Assign  (left  (Assign-left ex))
					  (right (replace id id_ref (Assign-right  ex)))))
   ((Abs?     ex)   (instantiate::Abs     (patt  (Abs-patt ex))
					  (body  (replace id id_ref (Abs-body      ex)))))
   ((App?     ex)   (instantiate::App     (fun   (replace id id_ref (App-fun ex)))
					  (arg   (replace id id_ref (App-arg ex)))))
   ((StructE? ex)   (instantiate::StructE (left  (replace id id_ref (StructE-left  ex)))
					  (right (replace id id_ref (StructE-right ex)))))
   ((Ref?     ex)   (instantiate::Ref     (expr  (replace id id_ref (Ref-expr      ex)))))
   ((Deref?   ex)   (instantiate::Deref   (expr  (replace id id_ref (Deref-expr    ex)))))))
;;
;;
;;
;;
;; PARSER FUNCTION
;;
(define (parse input)
   (read/lalrp
    (lalr-grammar
    (
     symbol_exit symbol_theory symbol_clean
     (left:  symbol_bang symbol_ref)
     (left:  symbol_assign)
     (right: symbol_arrow)
     (right: symbol_comma)
     (right: symbol_def)
     (right: symbol_seq)
     (right: symbol_pipe)
     ident
     string
     numeric
     symbol_lpar  symbol_rpar
     symbol_lqpar symbol_rqpar
     )
    ;;
    (expression
     ((symbol_theory)     (if (eq? *cleaning_wrong* #t)
			      (begin 
				 (set! *cleaning_wrong* #f)
				 (instantiate::Msg  (msg "Switching to the Empty Theory")))
			      (begin
				 (set! *cleaning_wrong* #t)
				 (instantiate::Msg  (msg "Switching to the Nostuck Theory")))))
     ((symbol_clean)       (begin (init_namespace)
				  (instantiate::Msg  (msg "Cleaning the Namespace"))))
     ((symbol_exit)        (instantiate::Exit    (msg "Bye bye, hoping see you again ...")))
     ((ident)              (instantiate::Var     (name  ident)))
     ((string)             (instantiate::Alg     (value string)))
     ((numeric)            (instantiate::Alg     (value numeric)))
     ;; REF EXPR
     ((symbol_bang expression)  (instantiate::Deref   (expr expression)))
     ;; DEFEF EXPR
     ((symbol_ref expression)   (instantiate::Ref     (expr expression)))
     ;; ARROW EXPR
     ((expression@ptn
       symbol_arrow
       expression@bdy)     (instantiate::Abs     (patt ptn)  (body bdy)))
     ;; STRUCTURE EXPR
     ((expression@left
       symbol_comma
       expression@right)   (instantiate::StructE (left left) (right right)))
     ;; ASSIGN EXPR
     ((expression@left
       symbol_assign
       expression@right)   (instantiate::Assign  (left left) (right right)))
     ;; NAMESPACE EXPR
     ((ident
       symbol_def
       expression)         (begin 
			      (add_namespace ident expression)
			      (instantiate::Var (name ident))))
     ;; SEQUENCE EXPR
     ((expression@exp1
       symbol_seq
       expression@exp2)    (instantiate::App
			      (fun (instantiate::Abs
				      (patt (instantiate::Var (name (gensym 'FRESH))))
				      (body exp2)))
			      (arg exp1)))
     ;; ()-PARENTHESISED CURRYFIED EXPR
     ((symbol_lpar
       expression_list
       symbol_rpar)             (curry expression_list))
     ;; []-PARENTHESISED CURRYFIED EXPR (DEFINE A TRS)
     ((symbol_lqpar
       def_list
       symbol_rqpar)            def_list))
     ;; EXPRESSION LIST
     (expression_list
     (() '())
     ((expression
       expression_list)    (cons expression expression_list)))
     ;; DEF LIST
     (def_list
	(()                (instantiate::Msg (msg "Term Rewriting System Definition")))
	((ident
	  symbol_def
	  expression)      (begin 
			      (set!fix ident expression)
			      (instantiate::Msg (msg "Term Rewriting System Definition"))))
	((ident
	  symbol_def
	  expression
	  symbol_pipe
	  def_list)        (begin 
			      (set!fix ident expression)
			      def_list))))
     ;;
   (regular-grammar ()
      ((+ (or #\Tab #\Space #\Newline  blank)) (ignore))
      (";;"   'symbol_top)
      (";"    'symbol_seq)
      ("|"    'symbol_pipe)
      ("->"   'symbol_arrow)
      ("("    'symbol_lpar)
      (")"    'symbol_rpar)
      ("["    'symbol_lqpar)
      ("]"    'symbol_rqpar)
      (","    'symbol_comma)
      ("ref"  'symbol_ref)
      ("!"    'symbol_bang)
      (":="   'symbol_assign)
      ("="    'symbol_def)
      ("if"   'symbol_if)
      ("then" 'symbol_then)
      ("else" 'symbol_else)
      ("$"    'symbol_theory)
      ("#"    'symbol_clean)
      ("@"    'symbol_exit)
      ((: upper (* (or upper digit))) (cons  'ident   (string->symbol (the-string))))
      ((: lower (* (or lower digit))) (cons  'string  (string->symbol (the-string)))) 
      ((+ digit)                      (cons  'numeric (string->symbol (the-string))))
      (else     (raise "Strange Input")))
      input
      (lambda (input) (eq? input 'symbol_top))))


