;---------------------------------------------------------------------
;  	A Certified implementation of the                          
;	 	                                  
;	I M P E R A T I V E  R H O  C A L C U L U S                  
;                                   
;       Powered by Bigloo-Scheme and Coq                    
;                                   
;       Copyright (c) INRIA 2004-2005 
;       Luigi Liquori and Bernard Paul Serpette 
;                                   
;  	Bug descriptions, use reports,                         
;	comments or suggestions are                            
;  	welcome. Send them to                            
;	                                  
;	Luigi.Liquori@inria.fr                             
;       Bernard.Serpette@inria.fr                            
;                  
;Copyright (C) 2004-2005 Luigi Liquori, Bernard Paul Serpette, INRIA.
;
;This library is free software; you can redistribute it and/or
;modify it under the terms of the GNU Lesser General Public
;License as published by the Free Software Foundation; either
;version 2.1 of the License, or (at your option) any later version.
;
;This library is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;Lesser General Public License for more details.
;
;You should have received a copy of the GNU Lesser Ge`neral Public
;License along with this library; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;-----------------------------------------
;;
;;
;;
;;
;;
;; MODULE DECLARATIONS AND DATATYPES
;;
(module imprho
	(main interpreter)
	(static
	 (class Expr)
	 (class NExpr::Expr     expr::Expr)
	 (class Alg::Expr      value::symbol)
	 (class Var::Expr       name::symbol)
	 (class GVar::Expr      name::Var    guard::Expr)
	 (class PExpr::Expr     prot::Expr    hand::Abs)
	 (class RExpr::Expr     rais::Expr)
	 (class Abs::Expr       patt::Expr    body::Expr)
	 (class App::Expr        fun::Expr     arg::Expr)
	 (class StructE::Expr   left::Expr   right::Expr)
	 (class Ref::Expr       expr::Expr)
	 (class Deref::Expr     expr::Expr)
	 (class Assign::Expr    left::Expr   right::Expr)
	 (class Effect::Expr      msg::string)
	 (class Msg::Expr        msg::string)
	 (class Exit::Expr       msg::string)
;;
;;
	 (class Value)
	 (class NValue::Value  value::Value)
	 (class Lit::Value     value::symbol)
	 (class Call::Value      fun::Value    arg::Value)
	 (class Clos::Value    funct::Abs env)
	 (class StructV::Value  left::Value  right::Value)
	 (class Loc::Value      addr::int)
	 (class PValue::Value   prot::Value   hand::Abs)
	 (class RValue::Value   rais::Value)
	 (class EffectV::Value   msg::string)
	 (class MsgV::Value      msg::string)
	 (class ExitV::Value     msg::string)))
;;
;; LOADING NOSTUCK THEORY
;;
(define *cleaning_wrong*  #t)
;;                                   
;; READ EVAL PRINT LOOP                           
;;                              
(define
   (interpreter argv)   
   (begin
    (print " ----------------------------------------------")
    (print "|               ----------\\                    |")
    (print "|               | i R h o  >                   |")
    (print "|               ----------/                    |")
    (print "| An Imperative Rewriting Calculus Interpreter |")
    (print "| Kernel Certified by the Proof Assistant Coq  |")
    (print "|           Powered by Bigloo Scheme           |")
    (print "|             Copyright Inria 2005             |")
    (print "|             NoEffect Theory Loaded           |")
    (print "|               Version 1.1alpha               |")
    (print "|               ? = Learn iRho                 |")
    (print " ---------------------------------------------- ")
;
;
    (let loop ()
      (bind-exit (return)
		 (with-exception-handler
		  (lambda (exc) 
		    (if (&error? exc)
			(print "(Parse Error on " (&error-obj exc) ")")
			(print "(Exception Error on " exc ")"))
		    (return #t))
		  (lambda ()
		    (init_store)
		    (display* "iRho << ")
		    (let ((val (Eval (parse (current-input-port)) *namespace*)))
		      (print  "iRho >> " (extern val))))))
      (loop))))
;;
;;
;; DEFINE WRONG BY SEGMENTATION VIOLATION
;;
(define *segment_fault*
  (lambda (loc)
    (instantiate::EffectV
     (msg (string-append "Segmentation Violation on Location "
			 (integer->string (Loc-addr loc)))))))
;;
;;
;; DEFINE WRONG BY UNBOUND VARIABLE
;;
(define *unbound_var*
  (lambda (var)
    (instantiate::EffectV
     (msg (string-append "Unbound Variable "
			 (symbol->string (Var-name var)))))))
;;
;;
;; DEFINE A STORE AN ADDRESS_STORE_COUNTER AND AN INIT_STORE
;;
(define *store* *segment_fault*)
;;
(define *address_store_counter* 0)
;;
(define (init_store)
  (set! *address_store_counter* 0)
  (set! *store* *segment_fault*))
;;
;;
;;
;; DEFINE A NAMESPACE, AN INIT_NAMESPACE 
;;
(define *namespace* *unbound_var*)
;;
(define (init_namespace)
  (set! *namespace* *unbound_var*))
;;
;;
;; ADD AN IDENTIFIER IN A GLOBAL NAMESPACE
;;
(define (add_namespace ident expression)
  (let ((val     (Eval expression *namespace*))
	(namesp  *namespace*))
    (set! *namespace* (lambda (v) (if (eq? (Var-name v) ident) val (namesp v))))))
;;
;;
;;
;; CREATE NEW LOCATIONS IN A STORE 
;;
(define (new_location val::Value)
  (let ((newloc (instantiate::Loc (addr *address_store_counter*))))
    (set! *address_store_counter* (+ *address_store_counter* 1))
    (let ((old_store *store*))
      (set! *store* (lambda (loc) (if (eq? loc newloc) val (old_store loc)))))
    newloc))
;;
;;
;;
;;                                  
;; INPUT                              
;;                                  
(define-generic (code expr::Expr))
;;                          
(define-method (code expr::NExpr) 
  (list "\\" (code (NExpr-expr expr))))
;;
(define-method (code expr::PExpr)
   (list 'Protect (code (PExpr-prot expr)) 'Handler (code (PExpr-hand expr))))
;;
(define-method (code expr::RExpr)
   (list 'Raise (code (RExpr-rais expr))))
;   (list 'Raise (code (RExpr (RExpr-rais expr)))))
;;
(define-method (code expr::Alg) 
  (Alg-value expr))
;;
(define-method (code expr::Var)                         
  (Var-name expr))
;;
(define-method (code expr::GVar)                         
  (list (code (GVar-name expr)) "\\" (code (GVar-guard expr))))
;;
(define-method (code expr::Abs)
  (list 'Fun (code (Abs-patt expr)) '-> (code (Abs-body expr)) ))
;;
(define-method (code expr::App)                         
  (list (code (App-fun expr)) (code (App-arg expr))))
;;
(define-method (code expr::StructE)                         
  (list (code (StructE-left expr))
	","
	(code (StructE-right expr))))
;;                      
(define-method (code expr::Ref)                         
  (list 'Ref (code (Ref-expr expr))))
;;
(define-method (code expr::Deref)                         
  (list 'Bang (code (Deref-expr expr))))
;;
(define-method (code expr::Assign)                         
  (list 'Ass
	(code (Assign-left expr))
	(code (Assign-right expr))))
;;
(define-method (code expr::Effect)
  (Effect-msg  expr))
;
(define-method (code expr::Msg)
  (Msg-msg  expr))
;;
(define-method (code expr::Exit)
  (Exit-msg  expr))
;;
;;                          
;; OUTPUT                             
;;                                  
(define-generic (extern value::Value))                        
;;
(define-method (extern value::NValue)                         
  (list "\\" (extern (NValue-value value))))
;;
(define-method (extern value::PValue)
   (list 'Protect (extern (PValue-prot value))
	 'Handler (code   (PValue-hand value))))
;;
(define-method (extern value::RValue)
   (list 'Raise (extern (RValue-rais value))))
;;
(define-method (extern value::Lit)                         
  (Lit-value value))
;;                            
(define-method (extern value::Call)                          
  (let ((f (extern (Call-fun value)))
	(a (extern (Call-arg value))))
    (if (pair? f) (append f (cons a '())) (list f a))))
;;                                 
(define-method (extern value::Clos) 
  (code (Clos-funct value)))
;;                                  
(define-method (extern value::StructV)                         
  (list (extern (StructV-left value)) "," (extern (StructV-right value))))
;;                                  
(define-method (extern value::Loc)
  (list 'Ref  (extern (*store* value))))
;;
(define-method (extern value::EffectV)                         
  (string-append "(Effect: " (EffectV-msg value) ")"))
;;
(define-method (extern value::MsgV)                         
  (MsgV-msg value))
;;
(define-method (extern value::ExitV)                         
  (print (ExitV-msg value))
  (exit 0))
;;                                
;;
;;
;;
;;
;;
;; 
;; S T A R T  O F  T H E  C E R T I F I E D  C O D E
;;
;;
;;
;; EVAL FUNCTION                               
;;                                  
(define-generic (Eval::Value expr::Expr env))                              
;;
(define-method (Eval::Value expr::PExpr env)
   (bind-exit (return)
      (with-exception-handler
       (lambda (raiseval)
	  (let ((hand (Eval (PExpr-hand expr) env)))
	     (return (call hand (RValue-rais raiseval)))))  ; handle the exception
       (lambda () (Eval (PExpr-prot expr) env))))) ; evaluate the protected body
;;
(define-method (Eval::Value expr::RExpr env)
   (let ((rais (Eval (RExpr-rais expr) env)))
      (raise (instantiate::RValue (rais rais)))))
;;		
(define-method (Eval::Value expr::NExpr env)
  (let ((val (Eval (NExpr-expr expr) env)))
    (if (NValue? val) 
	(NValue-value val) 
	(instantiate::NValue (value val)))))
;;
(define-method (Eval::Value expr::Alg env)                         
  (instantiate::Lit (value (Alg-value expr))))
;;            
(define-method (Eval::Value expr::Var env)                       
  (let ((envalue (env expr)))
    (if (EffectV? envalue) (*namespace* expr) envalue))) ; look first in the env and then in the namespace
;;     
(define-method (Eval::Value expr::GVar env)                       
  (let ((envalue  (Eval (GVar-name  expr) env))
	(genvalue (Eval (GVar-guard expr) env)))
     (cond ((EffectV? genvalue) envalue)               ; no guard 
	   ((not (equal? envalue genvalue)) envalue)   ; guard ok
	   (else (instantiate::EffectV (msg "Guard Lock"))))))
;; 
(define-method (Eval::Value expr::Abs env)                         
  (instantiate::Clos (funct expr) (env env)))
;;                                  
(define-method (Eval::Value expr::App env)                         
  (let   ((fun (Eval (App-fun expr) env)))
    (let ((arg (Eval (App-arg expr) env)))
      (call fun arg))))
;;                                  
(define-method (Eval::Value expr::StructE env)
   (let ((left  (Eval (StructE-left  expr) env))
	 (right (Eval (StructE-right expr) env))) ; evaluate left and right
      (if *cleaning_wrong*                        ; check the theory
	  (cond ((EffectV? left)  right)
		((EffectV? right) left)  
		(else (instantiate::StructV  (left left) (right right))))
	  (instantiate::StructV (left left) (right right)))))
;;       
(define-method (Eval::Value expr::Ref env)     
  (new_location (Eval (Ref-expr expr) env)))
;;                                  
(define-method (Eval::Value expr::Deref env)                         
  (let ((loc (Eval (Deref-expr expr) env)))
    (if (Loc? loc)
	(*store* loc)
	(instantiate::EffectV (msg "Deref is Not a Location")))))
;;                                  
(define-method (Eval::Value expr::Assign env)                         
  (let ((val (Eval (Assign-right expr) env)))
    (let ((loc (Eval (Assign-left expr) env)))
      (if (Loc? loc)
	  (let ((old *store*))
	    (set! *store* (lambda (x) (if (eq? x loc) val (old x))))
	    val)
	  (instantiate::EffectV (msg "Assign is Not a Location"))))))
;;    
(define-method (Eval::Value expr::Effect env)                         
  (instantiate::EffectV (msg (Effect-msg expr))))
;;
(define-method (Eval::Value expr::Msg env)                         
  (instantiate::MsgV (msg (Msg-msg expr))))
;;
(define-method (Eval::Value expr::Exit env)                         
  (instantiate::ExitV (msg (Exit-msg expr))))
;;                              
;;
;;
;;                   
;; CALL FUNCTION                                 
;;                                  
(define-generic (call::Value fun::Value arg::Value))                     
;;
(define-method (call::Value fun::Value arg::Value)  
  (instantiate::Call (fun fun) (arg arg)))
;;
(define-method (call::Value fun::Clos arg::Value)                         
  (let ((newenv (match (Abs-patt (Clos-funct fun)) arg (Clos-env fun))))
    (if newenv 
 	(Eval (Abs-body (Clos-funct fun))  newenv)
	(instantiate::EffectV (msg "Pattern Mismatch")))))
;;                          
(define-method (call::Value fun::StructV arg::Value)                         
   (let ((left  (call (StructV-left  fun) arg))
	 (right (call (StructV-right fun) arg))) ; evaluate left and right
      (if *cleaning_wrong*                       ; check the theory
	  (cond ((EffectV? left)  right)
		((EffectV? right) left)
		(else (instantiate::StructV (left left) (right right))))
	  (instantiate::StructV (left left) (right right)))))
;;                                  
(define-method (call::Value fun::Loc arg::Value)
  (instantiate::EffectV (msg "Calling a Location")))
;;                                  
(define-method (call::Value fun::EffectV arg::Value)                         
  fun)
;;    
;;
;;                              
;;
;;
;;
;;
;; MATCH FUNCTION
;;
(define-generic (match patt::Expr val::Value env))
;; 
(define-method (match patt::NExpr val::Value env)
   (if (NValue? val)
       (let ((newenv (match (NExpr-expr patt) (NValue-value val) env)))
	  (if newenv newenv env))
       (let ((newenv (match (NExpr-expr patt) val env)))
	  (if newenv #f env))))
;;                              
(define-method (match patt::Alg val::Value env)  
  (if (and (Lit? val) 
	   (equal? (Lit-value val) (Alg-value patt)))
      env
      #f))
;;
(define-method (match patt::Var val::Value env)
  (let ((envalue (env patt))) 
    (cond
     ((EffectV? envalue)     (lambda (x) (if (eq? (Var-name x) (Var-name patt)) val (env x))))
     ((equal? envalue val)  env)
     (else                  #f))))
;;
(define-method (match patt::GVar val::Value env)
  (let ((envalue  (env  (GVar-name  patt)))       ; bind variable in the env
	(genvalue (Eval (GVar-guard patt) env)))  ; eval guard in the same env
     (cond
	((and (EffectV? envalue)     (EffectV? genvalue)) 
	 (lambda (x) (if (eq? (Var-name x) (Var-name (GVar-name patt))) val (env x)))) ; no binding & no guard
	((and (EffectV? envalue)     (not (equal? val genvalue))) 
	 (lambda (x) (if (eq? (Var-name x) (Var-name (GVar-name patt))) val (env x)))) ; no binding & guard ok
	((and (equal? envalue val) (not (equal? envalue genvalue))) env)               ;    binding & guard ok
	((and (equal? envalue val) (EffectV? genvalue)) env)                           ;    binding & no guard
	(else    (print "     _Warning: Guard Lock_ ")            #f))))               ;              guard ko
;;
(define-method (match patt::App val::Value env)
  (cond
   ((Var? (App-fun patt))  (print "     _Warning: Active Variable_ ") #f)
   ((Call? val)
    (let ((newenv (match (App-fun patt) (Call-fun val) env)))
      (if  newenv (match (App-arg patt) (Call-arg val) newenv) #f)))
   (else  #f)))
;;              
(define-method (match patt::StructE val::Value env)   
  (if (StructV? val)
      (let ((newenv (match (StructE-left  patt) (StructV-left  val) env)))
	(if newenv (match (StructE-right patt) (StructV-right val) newenv) #f))
      #f))
;;           
(define-method (match patt::Ref val::Value env)                         
  (if (Loc? val)
      (match (Ref-expr patt) (*store* val) env)
      #f))
;;                                  
(define-method (match patt::Deref val::Value env)                         
  #f)
;;                                  
(define-method (match patt::Assign val::Value env)                         
  #f)
;;        
(define-method (match patt::Abs val::Value env) 
  #f)           
;;        
;;
;;
;;
;; 
;; E N D  O F  T H E  C E R T I F I E D  C O D E
;;
;;
;;
;;
;;
;; THE PARSER                               
;;             
;;
;; CURRIFICATION FUNCTIONS
;;
(define (curry list)
  (if (null? list) 
      (instantiate::Alg (value list))
      (curry-aux (car list) (cdr list))))
;;  
(define (curry-aux fun list)
  (if (null? list)
      fun
      (curry-aux (instantiate::App (fun fun) (arg (car list))) (cdr list))))
;;
;;
;; PARSER FUNCTION
;;
(define (parse input)
  (read/lalrp
   (lalr-grammar
    (
     symbol_exit symbol_theory symbol_clean symbol_help
     (left:  symbol_bang symbol_ref symbol_not symbol_raise_exception)
     (left:  symbol_assign)
     (right: symbol_rarrow_stop)
     (right: symbol_rarrow)
     (right: symbol_comma)
     (right: symbol_on_exception)
     (right: symbol_def)
     (right: symbol_seq)
     (right: symbol_pipe)
     (left:  symbol_larrow_else)
     (right: symbol_rarrow_then)
     ident
     string
     numeric
     symbol_lpar  symbol_rpar
     symbol_lqpar symbol_rqpar
     )
    ;;
    (expression
     ;; SYMBOL THEORY
     ((symbol_theory)     (if (eq? *cleaning_wrong* #t)
			      (begin 
			       (set! *cleaning_wrong* #f)
			       (instantiate::Msg  (msg "Switching to the Empty Theory")))
			    (begin
			     (set! *cleaning_wrong* #t)
			     (instantiate::Msg  (msg "Switching to the Nostuck Theory")))))
     ;; SYMBOL CLEAN
     ((symbol_clean)       (begin (init_namespace)
				  (instantiate::Msg  (msg "Cleaning the Namespace"))))
     ;; SYMBOL HELP
     ((symbol_help)        (begin
			    (print "")
			    (print "")
			    (print " <Non_Cap_alphadgt> as constants                       e.g., a,b,x,y,foo2 ")
			    (print " <Cap_alphadgt>     as variables                       e.g., A,B,X,Y,foo2 ")
			    (print "")
			    (print " ;;      as the Symbol of Top Level ") 
			    (print "")
 			    (print " ,       as the Symbol of Structures                   e.g. (3,X,foo) ")
			    (print " ->      as the Symbol of Function                     e.g. X->X ")
			    (print " ->|     as the Symbol of Resultless Function          e.g. X->| ")
			    (print "")
			    (print " <Space> as the Symbol of Application, e.g. (X->X 4) ") 
			    (print "")
			    (print " ^       as the Symbol of Create a Pointer to Data     e.g. ^X ")
			    (print " !       as the Symbol of Access Data via a Pointers   e.g. !X ")
			    (print " <Expr1> <- <Expr2> Assign <Expr2> to <Expr1>          e.g. X<-0 ")
			    (print "")
			    (print " [<Expr> | ...| <Expr>] define a Term Rewriting System ")
			    (print "")
			    (print " <Expr1>  ; <Expr2> Evaluate first <Expr1> and then <Expr2> ")
			    (print " <ElseExpr> <-? <CondExpr> ?-> <ThenExpr>  Conditional ")
			    (print " <HandExpr>  -) <ProtExpr>                 Declare Exception ")
			    (print "                           (-  <RaiseExpr> Raise   Exception ")
			    (print "")
			    (print " <Cap_alphadgt> = <Expr> bind <Expr> to <Cap_alphadgt>  e.g. X = 0 ")
			    (print "")
			    (print " $ to Switch Theory (Noeffect vs Syntactical)")
			    (print " # to Clean Namespace ")
			    (print " @ to Exit iRho ")			   
			    (print "")
			    (print "")
			    (instantiate::Msg  (msg "Please Visit www-sop.inria.fr/mirho/Luigi.Liquori/iRho/"))
			    ))
     ((symbol_exit)        (instantiate::Exit    (msg "Bye Bye Hoping See You Again ...")))
     ((ident)              (instantiate::Var     (name  ident)))
     ((string)             (instantiate::Alg     (value string)))
     ((numeric)            (instantiate::Alg     (value numeric)))
     ;; GUARD EXPR
     ((ident symbol_not expression)
                                (instantiate::GVar
				   (name (instantiate::Var (name ident)))
			           (guard expression)))
     ;; NOT EXPR
     ((symbol_not expression)   (instantiate::NExpr (expr expression)))
     ;; EXCEPTION EXPR
     ((expression@handler
       symbol_on_exception
       expression@protected)    (instantiate::PExpr (prot protected)
						    (hand handler)))
     ;; RAISEE EXPR
     ((symbol_raise_exception expression@raisee)
                                (instantiate::RExpr (rais raisee)))
     ;; REF EXPR
     ((symbol_bang expression)  (instantiate::Deref   (expr expression)))
     ;; DEFEF EXPR
     ((symbol_ref expression)   (instantiate::Ref     (expr expression)))
     ;; RARROW_STOP EXPR
     ((expression@ptn
       symbol_rarrow_stop)      (instantiate::Abs     (patt ptn)
						      (body (instantiate::Effect (msg "Monadic")))))
     ;; RARROW EXPR
     ((expression@ptn
       symbol_rarrow
       expression@bdy)          (instantiate::Abs     (patt ptn)  (body bdy)))
     ;; STRUCTURE EXPR
     ((expression@left
       symbol_comma
       expression@right)        (instantiate::StructE (left left) (right right)))
     ;; IFTHENELSE EXPR
     ((expression@else
       symbol_larrow_else
       expression@test
       symbol_rarrow_then
       expression@then)    (instantiate::App
			    (fun
			     (instantiate::StructE
			      (left (instantiate::Abs
				     (patt (instantiate::Alg (value 'false)))
				     (body
				      (instantiate::App
				       (fun
					(instantiate::Abs
					 (patt (instantiate::Var (name (gensym 'Fresh))))
					 (body else))
					)
				       (arg (instantiate::Alg (value 'dummy)))
				       )
				      )
				     )
				    )
			      (right (instantiate::Abs
				      (patt (instantiate::Alg (value 'true)))
				      (body
				       (instantiate::App
					(fun
					 (instantiate::Abs
					  (patt (instantiate::Var (name (gensym 'Fresh))))
					  (body then))
					 )
					(arg (instantiate::Alg (value 'dummy)))
					)
				       )
				      )
				     )
			      )
			     )
			    (arg test)))
     ;; ASSIGN EXPR
     ((expression@left
       symbol_assign
       expression@right)   (instantiate::Assign  (left left) (right right)))
     ;; NAMESPACE EXPR
     ((ident
       symbol_def
       expression)         (begin 
			     (add_namespace ident expression)
			     (instantiate::Var (name ident))))
     ;; SEQUENCE EXPR
     ((expression@exp1
       symbol_seq
       expression@exp2)    (instantiate::App
			    (fun (instantiate::Abs
				  (patt (instantiate::Var (name (gensym 'FRESH))))
				  (body exp2)))
			    (arg exp1)))
     ;; ()-PARENTHESISED CURRYFIED EXPR
     ((symbol_lpar
       expression_list
       symbol_rpar)        (curry expression_list))
     ;; []-PARENTHESISED CURRYFIED EXPR (DEFINE A TRS)
     ((symbol_lqpar
       trs_list
       symbol_rqpar)       trs_list))
     ;;
     ;; EXPRESSION LIST
    (expression_list
     (() '()) 
     ((expression
       expression_list)    (cons expression expression_list)))
     ;;
     ;; TRS LIST
    (trs_list
     (()                   (instantiate::Msg (msg "Empty Term Rewriting System Definition")))
     ((ident
       symbol_def
       expression)         (begin
			     (add_namespace ident expression)
			     (instantiate::Msg (msg "Term Rewriting System Definition"))))     
     ((ident
       symbol_def
       expression
       symbol_pipe
       trs_list)         (begin 
			   (add_namespace ident expression)
			   trs_list)))
    )
   (regular-grammar ()
		    ((+ (or #\Tab #\Space #\Newline blank)) (ignore))
		    (";;"   'symbol_top)
		    (";"    'symbol_seq)
		    ("|"    'symbol_pipe)
		    ("?->"  'symbol_rarrow_then)
		    ("<-?"  'symbol_larrow_else)
		    ("->"   'symbol_rarrow)
		    ("->|"  'symbol_rarrow_stop)
		    ("("    'symbol_lpar)
		    (")"    'symbol_rpar)
		    ("["    'symbol_lqpar)
		    ("]"    'symbol_rqpar)
		    (","    'symbol_comma)
		    ("^"    'symbol_ref)
		    (":-)"  'symbol_on_exception)
		    ("(-:"  'symbol_raise_exception)
		    ("!"    'symbol_bang)
		    ("\\"   'symbol_not)
		    ("<-"   'symbol_assign)
		    ("="    'symbol_def)
		    ("$"    'symbol_theory)
		    ("#"    'symbol_clean)
		    ("?"    'symbol_help)
		    ("@"    'symbol_exit)
		    ((: upper (* (or upper digit))) (cons  'ident   (string->symbol (the-string))))
		    ((: lower (* (or lower digit))) (cons  'string  (string->symbol (the-string)))) 
		    ((+ digit)                      (cons  'numeric (string->symbol (the-string))))
		    (else     (raise "Strange Input")))
   input
   (lambda (input) (eq? input 'symbol_top))))

